/*
 * This file is part of ascad.
 *
 * Ascad is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Ascad is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ascad; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
/* Copyright (C) 2005 Andrei Krger */

#include <stdio.h>
#include <stdarg.h>
#include "list.h"

typedef	unsigned char byte;
typedef	unsigned short word;
typedef	unsigned long dword;

typedef struct _encoder {
  int (*encode) (struct _encoder *, byte *data, long count);
  int (*flush) (struct _encoder *);
} encoder;

#define FLAG_RUSKETA 1
#define FLAG_POISTA_KAYRAT 2 /* Karttapaikan peruskartta */
#define FLAG_KARTTAPAIKAN_8000 4 /* Tee 8000->20000 muuttamalla tarkkuutta (n.)72dpi -> 245dpi */
#define FLAG_KVANTISOI 8 /* 0..127 -> musta, 127..253 -> harmaa, 253..254 -> valkoinen */

#define LOCATION_UPPER 1
#define LOCATION_RIGHT 2
#define LOCATION_LOWER_LEFT 0
#define LOCATION_UPPER_LEFT 1
#define LOCATION_LOWER_RIGHT 2
#define LOCATION_UPPER_RIGHT 3
#define LOCATION_CENTER 4

typedef struct {double x,y; struct list_head list;} point;
typedef struct {double x1,y1,x2,y2;} rect;
typedef struct {long x1,y1,x2,y2;} irect;

typedef struct {
  FILE *fp;
  long start;
  long length;
} file_block_t;

#define COL_TRANS_UNBROWN 0
#define COL_TRANS_COLORIZE 1 /* vrit vrill r/g/b, osuudella density */
#define COL_TRANS_CONTRAST 2 /* paranna kontrastia (maksimiin) */

typedef struct {
  struct list_head list;
  int mode;
  long r,g,b;
  double density; /* 0-100 */
} color_transform_t;

typedef struct {
  struct list_head list;
  char *fn;
  char *config;
  long mtime;
  double resolution; /* dpi, 0 = use image resolution or 72 */
  byte  location_type;
  point location;
  int   crop_length; /* -1 = no crop, 0 = rect crop, > 2 crop polygon */
  struct list_head crop_path; /* 72dpi in original source */
  rect  crop_box;
  double rotate;
  double scale; /* 1.00 */
  struct list_head col_trans_list; /* color_transform_t */
  file_block_t encoded;
} source_t;

typedef struct {
  struct list_head list;
  char *fn;
  char *config;
} ocad_t;

typedef struct {
  struct list_head list;
  char *fn;
  char *config;
} result_t;

#define HASH_LEN (2*16)

typedef struct {
  struct list_head list;
  char *name;
  char *config;
  char hash[HASH_LEN+1];
  rect page;
  double resolution;
  struct list_head source_list; /* source_t */
  struct list_head ocad_list; /* ocad_t */
  struct list_head result_list; /* resut_t */
} map_t;

typedef struct {
  long cols;
  long rows;
  double res_x; /* dots / 72dpi-pixel */
  double res_y; /* dots / 72dpi-pixel */
  byte **row_table;
} image_data_t;

#define min(a,b) ((a)<=(b) ? (a) : (b))
#define max(a,b) ((a)>=(b) ? (a) : (b))

#define MM2PT 720.0/254.0
#define PI 3.1415926535897932384626433832795029L 
#define DEG2RAD PI/180

#ifdef ASCAD_MAIN
#define EXT 
#else
#define EXT extern
#endif

EXT void DEB(const char *fmt, ...);
EXT void xperror(const char *fmt, ...);
EXT void xerror(const char *fmt, ...);
EXT void *xmalloc(size_t size);
EXT char *dospath;
EXT char *errstr;

encoder * InitA85Encoder(FILE *out);
encoder * InitGZEncoder(encoder *ec);
int encode_source(FILE *out, source_t *source, rect *page);
int load_image(FILE *fp, image_data_t *img);


