/*
 * This file is part of ascad.
 *
 * Ascad is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Ascad is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ascad; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
/* Copyright (C) 2005 Andrei Krger */

#include <stdio.h>
#include <stdarg.h> //vfprintf, etc.
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <sys/stat.h>
#include <math.h>

#include "ascad.h"

int ps_comment(char *line, int *state) {
  if (*state) {
    if (strncmp(line, "%%DataEnd", 9) == 0) {
      *state = 0;
    }
    return 0;
  }
  if (strncmp(line, "%%DataBegin", 11) == 0) {
    *state |= 1;
    return 0;
  }
  if (strncmp(line, "%%", 2) == 0 ||
      strncmp(line, "%!", 2) == 0 ||
      strncmp(line, "showpage", 8) == 0) {
    return 1;
  }
  return 0;
}

#define LINE_SIZE 1024

int ispostscript(FILE *f)
{
  byte buf[4];

  fread(buf, 1, 4, f);
  rewind(f);
  if (strncmp(buf, "%!PS", 4) == 0)
    return 1;
  return 0;
}

int read_boundingbox(FILE *fp, rect *box) 
{
  int i;
  byte line[LINE_SIZE];

  for(i=0;i<50;i++) 
  {
    if (fgets(line, LINE_SIZE, fp) < 0) {i=50; break;}
    if (sscanf(line, "%%%%BoundingBox: %lf %lf %lf %lf", &box->x1, &box->y1, &box->x2, &box->y2) == 4)
      break;
  }
  if (fseek(fp, 0, SEEK_SET)) return -1;
  return (i == 50);
}

int process_ps(FILE *out, FILE *fp, source_t *source, rect *page) { 
  byte line[LINE_SIZE];
  point origo;
  int ps_state;

  fprintf(out, "save 9 dict begin %.3f %.3f translate %.3f rotate %.3f %.3f scale\n",
      source->location.x, source->location.y, -(source->rotate), source->scale, source->scale);

  if (source->location_type == LOCATION_LOWER_LEFT)
  {
    origo.x = (source->crop_length > 0 ? source->crop_box.x1 : 0.0);
    origo.y = (source->crop_length > 0 ? source->crop_box.y1 : 0.0);
  } 
  else
  {
    rect bb;
    if (read_boundingbox(fp, &bb) != 0) 
    {
      if (source->crop_length == 0)
        xerror("Error read BoundingBox (and no crop defined in case not 'lounas=')");
      bb.x1 = source->crop_box.x1;
      bb.y1 = source->crop_box.y1;
      bb.x2 = source->crop_box.x2;
      bb.y2 = source->crop_box.y2;
    }
    else
    {
      if (source->crop_length > 0)
      {
        bb.x1 = max(bb.x1, source->crop_box.x1);
        bb.y1 = max(bb.y1, source->crop_box.y1);
        bb.x2 = min(bb.x2, source->crop_box.x2);
        bb.y2 = min(bb.y2, source->crop_box.y2);
      }
    }

    origo.x = (source->location_type & LOCATION_RIGHT) ? bb.x2 : bb.x1;
    origo.y = (source->location_type & LOCATION_UPPER) ? bb.y2 : bb.y1;
    if (source->location_type == LOCATION_CENTER) {
      /* ANc 4.8.05: korjattu keskiarvon laskenta */
      origo.x = (bb.x2 + bb.x1) / 2;
      origo.y = (bb.y2 + bb.y1) / 2;
      /*****
      origo.x = (bb.x2 - bb.x1) / 2;
      origo.y = (bb.y2 - bb.y1) / 2;
      *****/
    }
  }

  if (origo.x || origo.y) /* move to location_type/crop */
    fprintf(out, "%.3f %.3f translate\n", -(origo.x), -(origo.y));

  if (source->crop_length > 0)
  {
    if (source->crop_length > 2)
    {
      int i=0;
      point *pt;

      list_for_each_entry(pt, &source->crop_path, list)
        fprintf(out, "%.3f %.3f %s ", pt->x, pt->y, (i++ ? "lineto" : "moveto"));
      fprintf(out, "closepath clip newpath\n");
    }
    else /* rect crop */
    {
      fprintf(out, "%.3f %.3f moveto %.3f %.3f lineto %.3f %.3f lineto %.3f %.3f lineto closepath clip newpath\n",
          source->crop_box.x1, source->crop_box.y1,
          source->crop_box.x1, source->crop_box.y2,
          source->crop_box.x2, source->crop_box.y2,
          source->crop_box.x2, source->crop_box.y1);
    }

  }

  if (source->scale != 1.0) 
  {
    /* Fix line width: 50% zoom -> 71% linewidth, 200% zoom --> 141% linewidth */
    fprintf(out, "/orig__setlinewidth /setlinewidth load def\n");
    fprintf(out, "/setlinewidth {%.3f mul orig__setlinewidth} def\n", sqrt(1.0/source->scale));
  }


  ps_state = 0;
  while (fgets(line, LINE_SIZE, fp)) {
    if (ps_comment(line, &ps_state)) fputs("%*", out);
    fputs(line, out);
  }

  fprintf(out, "end restore\n");

  return 0;
}

void rotate(point *p, point origo, double deg) {
  double ax = (p->x - origo.x)*cos(deg*DEG2RAD) + (p->y - origo.y)*sin(deg*DEG2RAD);
  double ay = (p->y - origo.y)*cos(deg*DEG2RAD) - (p->x - origo.x)*sin(deg*DEG2RAD);
  p->x = ax+origo.x; p->y = ay+origo.y;
}

#define get_intersect(a, min, max) ((a)>(max)?8:((a)>=(min)?4:2))
#define y_intersect(p1,p2,x0) (p2.y-p1.y)/(p2.x-p1.x)*(x0-p1.x) + p1.y
#define x_intersect(p1,p2,y0) (p2.x-p1.x)/(p2.y-p1.y)*(y0-p1.y) + p1.x

int clip_region(point region[] /* 4 points in, 0-8 points out */, rect crop)
/* Sutherland-Hodgman like Polygon Clipper */
{
  int i,j, n,c;
  byte intersect;
  point temp[6];


  /* clip horizontal (x) */

  n = 4;
  intersect = get_intersect(region[n-1].x, crop.x1, crop.x2);

  for(i=0, j=n-1, c=0; i<n; j=i++) {
    intersect = ((intersect<<4) & 0xf0) | get_intersect(region[i].x, crop.x1, crop.x2);

    if (intersect < 0x88 && intersect > 0x22) {
      if (intersect > 0x44) {
        temp[c].x = crop.x2; // add Right (max) intersection
        temp[c++].y = y_intersect(region[j], region[i], crop.x2);
        if (intersect == 0x82) {
          temp[c].x = crop.x1; // add Left (min) intersection
          temp[c++].y =  y_intersect(region[j], region[i], crop.x1);
        }
      }
      else if (intersect & 0x22) {
        temp[c].x = crop.x1; // add Left (min) intersection
        temp[c++].y =  y_intersect(region[j], region[i], crop.x1);
        if (intersect == 0x28) {
          temp[c].x = crop.x2; // add Right (max) intersection
          temp[c++].y = y_intersect(region[j], region[i], crop.x2);
        }
      }
      if (intersect & 0x04) {
        temp[c].x = region[i].x; // add End point of the segment
        temp[c++].y = region[i].y;
      }
    }
  }

  /* clip vertically (y) */

  n = c;
  intersect = get_intersect(temp[n-1].y, crop.x1, crop.x2);

  for(i=0, j=n-1, c=0; i<n; j=i++) {
    intersect = ((intersect<<4) & 0xf0) | get_intersect(temp[i].y, crop.y1, crop.y2);

    if (intersect < 0x88 && intersect > 0x22) {
      if (intersect > 0x44) {
        region[c].y = crop.y2; // add Top intersection
        region[c++].x = x_intersect(temp[j], temp[i], crop.y2);
        if (intersect == 0x82) {
          region[c].y = crop.y1; // add Bottom intersection
          region[c++].x =  x_intersect(temp[j], temp[i], crop.y1);
        }
      }
      else if (intersect & 0x22) {
        region[c].y = crop.y1; // add Bottom intersection
        region[c++].x =  x_intersect(temp[j], temp[i], crop.y1);
        if (intersect == 0x28) {
          region[c].y = crop.y2; // add Top intersection
          region[c++].x = x_intersect(temp[j], temp[i], crop.y2);
        }
      }
      if (intersect & 0x04) {
        region[c].y = temp[i].y; // add End point of the segment
        region[c++].x = temp[i].x;
      }
    }
  }

  return c;
}

int calculate_range(image_data_t *img, source_t *source, rect *page, irect *img_box, rect *drawing_box /*72dpi*/) {
  /* No huh.. Tm optiomoi kuvan rajauksen (img_box) niin, ett kuvamuunnokset 
   * voidaan tehd vain teholliselle alueelle. Drawing_box on kuvan piirtoalue
   * source->location kohdassa valmiiksi kierretylle koordinaatistolle.
   */
  rect base_crop; /* source->crop_box'ed area in source 72dpi coordinates */
  rect disp_crop; /* base_crop + page-crop (source 72dpi coordinates) */ 
  rect page_proj; /* projection of the page boundaries in source 72dpi coordinates */
  point origo; /* in source 72dpi coordinates  */
  long tmp;

  /* crop box in image 72dpi-coordinates */

  if (source->crop_length == 0) { /* no crop */
    base_crop.x1 = 0;
    base_crop.y1 = 0;
    base_crop.x2 = (double)img->cols / img->res_x; 
    base_crop.y2 = (double)img->rows / img->res_y;
  } else {
    base_crop.x1 = max(source->crop_box.x1, 0);
    base_crop.y1 = max(source->crop_box.y1, 0);
    base_crop.x2 = min(source->crop_box.x2, (double)img->cols / img->res_x);
    base_crop.y2 = min(source->crop_box.y2, (double)img->rows / img->res_y);
  }

  /* Calculate rotation point */

  origo.x = (source->location_type & LOCATION_RIGHT) ? base_crop.x2 : base_crop.x1;
  origo.y = (source->location_type & LOCATION_UPPER) ? base_crop.y2 : base_crop.y1;
  if (source->location_type == LOCATION_CENTER) {
    /* ANc 4.8.05: Korjattu keskiarvon laskenta */
    origo.x = (base_crop.x2 + base_crop.x1) / 2;
    origo.y = (base_crop.y2 + base_crop.y1) / 2;
    /*******
    origo.x = (base_crop.x2 - base_crop.x1) / 2;
    origo.y = (base_crop.y2 - base_crop.y1) / 2;
    *******/
  }

  /* transform, scale and rotate page box (all 4 corners) */

  page_proj.x1 = (page->x1 - source->location.x) / source->scale + origo.x;
  page_proj.y1 = (page->y1 - source->location.y) / source->scale + origo.y;
  page_proj.x2 = (page->x2 - source->location.x) / source->scale + origo.x;
  page_proj.y2 = (page->y2 - source->location.y) / source->scale + origo.y;

  if (source->rotate == 0.0)
  {
    /* crop image into page boundaries (easy case when rotate = 0) */
    disp_crop.x1 = max(base_crop.x1, page_proj.x1);
    disp_crop.y1 = max(base_crop.y1, page_proj.y1);
    disp_crop.x2 = min(base_crop.x2, page_proj.x2);
    disp_crop.y2 = min(base_crop.y2, page_proj.y2);
  }
  else 
  {
    /* rotate the page and clip it (!) into the image box */
    point region[8];
    int pcount, i;

    region[0].x = page_proj.x1; region[0].y = page_proj.y1;
    region[1].x = page_proj.x1; region[1].y = page_proj.y2;
    region[2].x = page_proj.x2; region[2].y = page_proj.y2;
    region[3].x = page_proj.x2; region[3].y = page_proj.y1;

    rotate(&region[0], origo, -(source->rotate));
    rotate(&region[1], origo, -(source->rotate));
    rotate(&region[2], origo, -(source->rotate));
    rotate(&region[3], origo, -(source->rotate));

    pcount = clip_region(region, base_crop); /* 4 points in, 0-8 points out */

    disp_crop.x1 = base_crop.x2; disp_crop.y1 = base_crop.y2; // initialize max
    disp_crop.x2 = base_crop.x1; disp_crop.y2 = base_crop.y1; // initialize min

    for(i=0; i<pcount; i++) {
      disp_crop.x1 = min(disp_crop.x1, region[i].x);
      disp_crop.y1 = min(disp_crop.y1, region[i].y);
      disp_crop.x2 = max(disp_crop.x2, region[i].x);
      disp_crop.y2 = max(disp_crop.y2, region[i].y);
    }
  }
   
  /* round to real pixels, always equal or greater, and recalc crop_box */

  img_box->x1 = floor(disp_crop.x1 * img->res_x + 0.001);
  img_box->y1 = floor(disp_crop.y1 * img->res_y + 0.001);
  img_box->x2 = ceil(disp_crop.x2 * img->res_x - 0.001);
  img_box->y2 = ceil(disp_crop.y2 * img->res_y - 0.001);

  disp_crop.x1 = img_box->x1 / img->res_x;
  disp_crop.y1 = img_box->y1 / img->res_y;
  disp_crop.x2 = img_box->x2 / img->res_x;
  disp_crop.y2 = img_box->y2 / img->res_y;

  /* do we have anything left? */
  if (!(img_box->x2 - img_box->x1 > 0 && img_box->y2 - img_box->y1 > 0))
    return -1;

  assert(img_box->x1 >=0);
  assert(img_box->y1 >=0);
  assert(img_box->x2 <= img->cols);
  assert(img_box->y2 <= img->rows);

  /* calculate drawing box */
  drawing_box->x1 = (disp_crop.x1 - origo.x) * source->scale;
  drawing_box->y1 = (disp_crop.y1 - origo.y) * source->scale;
  drawing_box->x2 = (disp_crop.x2 - origo.x) * source->scale;
  drawing_box->y2 = (disp_crop.y2 - origo.y) * source->scale;

  /* reverse crop_box top-to-bottom */
  tmp = img_box->y2;
  img_box->y2 = img->rows - img_box->y1;
  img_box->y1 = img->rows - tmp;

  return 0;
}

struct palette {
  union {
    struct { byte r,g,b; } rgb;
    dword key;
  } u;
  int next; /* link to next color of same hash */
};

int palette_full(byte **data, int x, int y, irect box, /*int cols,*/ struct palette *palette) {
  /* make fall back when colors count > 256 */
  for(; y>=box.y1; y--)
  {
    for(; x>=box.x1; x--)
    {
      byte id = data[y][x];
      data[y][x*3] = palette[id].u.rgb.r;
      data[y][x*3+1] = palette[id].u.rgb.g;
      data[y][x*3+2] = palette[id].u.rgb.b;
    }
    x = box.x2-1;
  }
  return -1;
}

int make_palette(byte **data, irect box, /*int rows, int cols,*/  struct palette *palette)
{
  int c=0;
  long x, y, i;
  struct palette hash_list[256];

  for(i=0; i<256; i++)
    hash_list[i].next = -1;
  
  for(y = box.y1; y < box.y2; y++)
  {
    for(x = box.x1; x < box.x2; x++)
    {
      struct palette color;
      byte hash;

      color.u.rgb.r =  data[y][x*3];
      color.u.rgb.g =  data[y][x*3+1];
      color.u.rgb.b =  data[y][x*3+2];

      hash = color.u.rgb.r ^ color.u.rgb.g ^ color.u.rgb.b;

      if ((i = hash_list[hash].next) < 0)
      {
        if (c==256)
          return palette_full(data, x-1, y, box, /*cols,*/ palette);

        hash_list[hash].next = c;
        palette[c].u.key = color.u.key;
        palette[c].next = -1;
        i = c++;
      }
      else
      {
        while(palette[i].u.key != color.u.key)
        {
          if (palette[i].next < 0)
          {
            if (c==256)
              return palette_full(data, x-1, y, box, /*cols,*/ palette);
            palette[i].next = c;
            palette[c].u.key = color.u.key;
            palette[c].next = -1;
            i = c++;
            break;
          }
          i = palette[i].next;
        }
      }

      data[y][x] = i;
    }
  }

  return c;
}

void bitcompress_row(byte *p, long length, int bits) {
  register int rot=0;
  byte *res = p;
  byte mask = (1<<bits)-1;

  assert (bits == 1 || bits == 2 || bits == 4);

  while(length--) {
    if (rot)
      *res |= ((*p++ & mask) << (rot -= bits));
    else 
      *res = ((*p++ & mask) << (rot += 8 - bits));
    if (!rot) res++;
  }
}

#define C(_r,_g,_b) ((_r<<16) | (_g<<8) | _b)

/* Poistaa Karttapaikan korkeuskyrien ruskean, aluperisen (0) ja varsinaiset (1).
 * Keltaisen (2) ja sinisen (3) pllevrjtyt alueet palautetaan alkuperisiksi. */
void unbrown(color_transform_t *col_trans, byte **data, irect box /*int rows, int cols*/) {
  long x=0, y;
  for(y = box.y1; y < box.y2; y++)
  {
    for(x = box.x1; x < box.x2; x++)
    {
      int col = (data[y][x*3]<<16) | (data[y][x*3+1]<<8) | data[y][x*3+2];
      if (col == C(171,87,0) ||
          col == C(192,129,63) || col == C(213,171,127) || col == C(234,213,191))
      { /* puhdas ruskea ja sen vaaleammat svyt valkoiseksi */
        data[y][x*3] = data[y][x*3+1] = data[y][x*3+2] = 255;
      }
      else if (col == C(192,129,38) || col == C(213,171,76) || col == C(234,213,114))
      { /* ruskean ja keltaisen sekoitus keltaiseksi */
        data[y][x*3] = data[y][x*3+1] = 255; data[y][x*3+2] = 153;
      }
      else if (col == C(157,129,63) || col == C(143,171,127) || col == C(129,213,191))
      { /* ruskean ja sinisen sekoitus siniseksi */
        data[y][x*3] = 115; data[y][x*3+1] = data[y][x*3+2] = 255;
      }
    }
  }
}

void set_contrast(color_transform_t *col_trans, byte **data, irect box /*int rows, int cols*/) {
  long _min=255, _max=0;
  long x, y;
  byte conv[256];

  for(y = box.y1; y < box.y2; y++)
    for(x = box.x1; x < box.x2*3; x++) 
    {
      _min = min(_min, data[y][x]);
      _max = max(_max, data[y][x]);
    }

  if (_min >= _max)
    return;

  for(x=0;x<256;x++)
    conv[x] = (x - _min) * 255 / (_max - _min);

  for(y = box.y1; y < box.y2; y++)
    for(x = box.x1; x < box.x2*3; x++) 
      data[y][x] = conv[data[y][x]];
}

void colorize(color_transform_t *col_trans, byte **data, irect box /*int rows, int cols*/) {
  long x, y;
  long r,g,b,Y;
  long R[256],G[256],B[256]; /* fast color lookups */
  long density = (long)col_trans->density;

  for(Y=0;Y<256;Y++)
  {
    R[Y] = col_trans->r + (Y*(255-col_trans->r)/255);
    G[Y] = col_trans->g + (Y*(255-col_trans->g)/255);
    B[Y] = col_trans->b + (Y*(255-col_trans->b)/255);
  }

  for(y = box.y1; y < box.y2; y++)
    for(x = box.x1; x < box.x2; x++) 
    {
      r = data[y][x*3];
      g = data[y][x*3+1];
      b = data[y][x*3+2];
      Y = (6969*r + 23434*g + 2365*b)/32768;
      data[y][x*3]   = (R[Y] * density + r * (100-density)) / 100;
      data[y][x*3+1] = (G[Y] * density + g * (100-density)) / 100;
      data[y][x*3+2] = (B[Y] * density + b * (100-density)) / 100;
    }
}

int process_image(FILE *out, FILE *fp, source_t *source, rect *page) {
  image_data_t img;
  rect drawing_box;
  irect img_box;
  long cols, rows;
  int bit_depth, channels;
  long row_start, row_size;
  color_transform_t *col_trans;
  int i;
  char *decode;
  encoder *ec[2];
  struct palette palette[256];
  int indexed;

  if (load_image(fp, &img)) /* always 24bit RGB RGB RGB format */
    return -1;

  if (source->resolution != 0)
    img.res_x = img.res_y = source->resolution/72;

  if (img.res_x == 0 || img.res_y == 0)
    img.res_x = img.res_y = 1.0; /* guess 72 dpi */

  if (calculate_range(&img, source, page, &img_box, &drawing_box)) {
    fprintf(stderr, "WARNING %s: out of page/crop\n", source->fn);
    return 0;
  }

  /* Vrimuunnokset */
  list_for_each_entry(col_trans, &source->col_trans_list, list)
  {
    switch(col_trans->mode) {
      case COL_TRANS_UNBROWN:
        unbrown(col_trans, img.row_table, img_box); break;
      case COL_TRANS_COLORIZE:
        colorize(col_trans, img.row_table, img_box); break;
      case COL_TRANS_CONTRAST:
        set_contrast(col_trans, img.row_table, img_box); break;
    }
  }

  /* Yrit luoda paletti (jos alle 256 svy) */
  indexed = make_palette(img.row_table, img_box, palette);

  /* WRITE postscript */

  fprintf(out, "save 9 dict begin %.3f %.3f translate %.3f rotate\n",
               source->location.x, source->location.y, -(source->rotate));

  if (source->crop_length > 2) /* crop region */
  {
    int i=0;
    point *pt, origo;

    origo.x = (source->location_type & LOCATION_RIGHT) ? source->crop_box.x2 : source->crop_box.x1;
    origo.y = (source->location_type & LOCATION_UPPER) ? source->crop_box.y2 : source->crop_box.y1;
    if (source->location_type == LOCATION_CENTER) {
      origo.x = (source->crop_box.x1 + source->crop_box.x2) / 2;
      origo.y = (source->crop_box.y1 + source->crop_box.y2) / 2;
    }

    fprintf(out, "newpath ");
    list_for_each_entry(pt, &source->crop_path, list)
      fprintf(out, "%.3f %.3f %s ", (pt->x - origo.x)*source->scale, (pt->y - origo.y)*source->scale, (!i++ ? "moveto" : "lineto"));
    fprintf(out, "closepath clip newpath\n");
  }

  if (indexed >= 0)
  {
    /* write palette */
    fprintf(out, "[/Indexed /DeviceRGB %d <", indexed-1);
    for(i=0; i<indexed; i++)
      fprintf(out, "%02X%02X%02X ", palette[i].u.rgb.r, palette[i].u.rgb.g, palette[i].u.rgb.b);
    fprintf(out, ">] setcolorspace\n");

    if (indexed > 15) { bit_depth=8; decode = "[0 255]";}
    else if (indexed > 3) { bit_depth=4; decode = "[0 15]";}
    else if (indexed > 1) { bit_depth=2; decode = "[0 3]";}
    else { bit_depth=1; decode = "[0 1]";}
    channels = 1;
  } 
  else
  {
    /* write rgb */
    fprintf(out, "/DeviceRGB setcolorspace\n");
    bit_depth = 8;
    channels = 3;
    decode = "[0 1 0 1 0 1]";
  }

  cols = img_box.x2 - img_box.x1;
  rows = img_box.y2 - img_box.y1;
  row_start = img_box.x1 * channels;
  row_size  = (cols*bit_depth*channels +7)/8;

  if (bit_depth < 8)
    for (i = img_box.y1; i < img_box.y2; i++)
      bitcompress_row(img.row_table[i] + row_start, cols, bit_depth);

  fprintf(out, "%.3f %.3f translate %.3f %.3f scale\n", drawing_box.x1, drawing_box.y1,
          drawing_box.x2 - drawing_box.x1, drawing_box.y2 - drawing_box.y1);
  fprintf(out, "{<< /ImageType 1 /Width %ld /Height %ld /BitsPerComponent %d\n", cols, rows, bit_depth);
  fprintf(out, "   /ImageMatrix [%ld 0 0 -%ld 0 %ld] /Decode %s\n", cols, rows, rows, decode);
  fprintf(out, "   /Interpolate true\n");
  fprintf(out, "   /DataSource currentfile /ASCII85Decode filter /FlateDecode filter\n");
  fprintf(out, ">> image}\n");
  fprintf(out, "%%%%DataBegin:\nexec\n");

  ec[0] = InitA85Encoder(out);
  ec[1] = InitGZEncoder(ec[0]);

  for (i = img_box.y1; i < img_box.y2; i++) {
    ec[1]->encode(ec[1], img.row_table[i] + row_start, row_size);
  }

  ec[1]->flush(ec[1]);

  fprintf(out, "%%%%DataEnd\nend restore\n");

  for(i=0; i<img.rows; i++)
    free(img.row_table[i]);

  return 0;
}


int encode_source(FILE *out, source_t *source, rect *page) {
  FILE *fp;

  if ((fp = fopen(source->fn, "r")) == NULL) 
    return -1;

  if (ispostscript(fp))
  {
    if (process_ps(out, fp, source, page))
      xperror("process_ps(%s)", source->fn);
  }
  else
  {
    if (process_image(out, fp, source, page))
      xperror("process_image(%s)", source->fn);
  }

  fclose(fp);
  return 0;
}

